@@ -1,14 +1,17 @@
 /*******************************************************************************
-**  fly:  On-the-fly GIF creation utility
-**  Martin Gleeson, ITS, gleeson@unimelb.edu.au
-**  Copyright (c), The University of Melbourne, 1994-1999
+**  fly:  On-the-fly image creation utility
+**  Martin Gleeson, martin@gleeson.com
+**  Copyright (c), Martin Gleeson and The University of Melbourne, 1994-2001
 **  Created: 9 December 1994
-**  Last Update: 15th January 1999
+**  Last Update: 28th February 2001
 **
 **  Uses the gd library by Thomas Boutell, boutell@boutell.com
-**  gd: Copyright 1994-1998, Quest Protein Database Centre, Cold Spring Harbour Labs
+**  gd: Copyright 1994-1999, Quest Protein Database Centre,
+**     Cold Spring Harbour Labs
 **
 **  Contributions from:
+**  Philip Warner
+**     patches to gd to retain support for GIF images
 **  Claus Hofmann <claush@ipfr.bau-verm.uni-karlsruhe.de>
 **     addition of 'transparent' directive.
 **     addtion of code to check if colour already allocated
@@ -60,16 +63,16 @@
 **  Strings for usage and quick reference
 ******************************************************************************/
 
-char *version = "1.6.5";
-char *usage = "Usage : fly [-h] [-q] [-i inputfile] [-o outputfile.gif]";
+char *version = "2.0.0";
+char *usage = "Usage : fly [-h] [-q] [-i inputfile] [-o outputfile]";
 
-char *help = "See <URL:http://www.unimelb.edu.au/fly/fly.html> for documentation.\n\
+char *help = "See <URL:http://martin.gleeson.com/fly/> for documentation.\n\
 \n\
 Quick Reference to Directives: \n\
 \n\
 new\n\
 size x,y\n\
-name filename.gif\n\
+name filename\n\
 \n\
 line    x1,y1,x2,y2,R,G,B             dline        x1,y1,x2,y2,R,G,B \n\
 rect    x1,y1,x2,y2,R,G,B             frect        x1,y1,x2,y2,R,G,B \n\
@@ -84,8 +87,8 @@ string   R,G,B,x,y,<size>,<string> \n\
 stringup R,G,B,x,y,<size>,<string> \n\
 (size = tiny, small, medium, large or giant) \n\
 \n\
-copy         x,y,x1,y1,x2,y2,filename.gif \n\
-copyresized  x1,y1,x2,y2,dx1,dy1,dx2,dy2,filename.gif \n\
+copy         x,y,x1,y1,x2,y2,filename \n\
+copyresized  x1,y1,x2,y2,dx1,dy1,dx2,dy2,filename \n\
 \n\
 setpixel     x,y,R,G,B \n\
 getpixel     x,y \n\
@@ -93,13 +96,16 @@ colourchange R1,G1,B1,R2,G2,B2 \n\
 transparent  R,G,B \n\
 interlace \n\
 \n\
-setbrush    filename.gif                         killbrush \n\
-settile     filename.gif                         killtile \n\
+setbrush    filename                         killbrush \n\
+settile     filename                         killtile \n\
 setstyle    R1,G1,B1,R2,G2,B2,...,Rn,Bn,Gn       killstyle \n\n\
 sizex \n\
 sizey \n\
 rotate      deg \n\
 \n\
+type        gif|png|jpeg\n\
+quality     n (image quality when type = jpeg. Must be in the range 0-95)\n\
+\n\
 end\n";
 
 /******************************************************************************
@@ -109,10 +115,10 @@ end\n";
 int         process_args(int argc, char *argv[]);
 int         get_token(FILE *infile);
 int         get_number(FILE *infile);
-char       *get_string(FILE *infile);
+char        *get_string(FILE *infile);
 void        sync_input(FILE *infile);
 int         get_colour(FILE *infile, gdImagePtr img);
-void        copy_to_gif(FILE *infile, gdImagePtr img, int resize);
+void        copy_to_img(FILE *infile, gdImagePtr img, int resize);
 gdImagePtr  get_image(int type, int argc, char *argv[]);
 void       *my_newmem(size_t size);
 FILE       *get_file(int *line_number, char **filename);
@@ -127,6 +133,8 @@ int   finished = FALSE,
          quiet = FALSE,
    end_of_line = FALSE,
    written_out = FALSE,
+          type = 0,
+       imgtype = PNG,
    line_number = 1;
 
 char     *input_file = NULL,
@@ -144,7 +152,7 @@ FILE     *verbose_out;
 
 int main(int argc, char *argv[]){
 	int colour = 0, colour1 = 0, colour2 = 0, type;
-	int status, size, side;
+	int status, size, side, jp_quality = 75;
 	int brush_on = 0, tile_on = 0, style_on = 0;
 	int num_entries, up = 0;
 	int i, n, c, x, y;
@@ -661,7 +669,7 @@ START:	type = get_token(infile);
 			}
 			else
 			{
-				brush = gdImageCreateFromGif(brushfile);
+				brush = gdImageCreateFromPng(brushfile);
 				gdImageSetBrush(img,brush);
 				brush_on = 1;
 				if (!quiet) fprintf(verbose_out,"## Brush Set ## to: %s\n",s);
@@ -703,7 +711,7 @@ START:	type = get_token(infile);
 			}
 			else
 			{
-				tile = gdImageCreateFromGif(tilefile);
+				tile = gdImageCreateFromPng(tilefile);
 				gdImageSetTile(img,tile);
 				tile_on = TRUE;
 				if (!quiet) fprintf(verbose_out,"## Tile Set ## to: %s\n",s);
@@ -716,11 +724,11 @@ START:	type = get_token(infile);
 			break;
 
 		case COPY:
-			copy_to_gif(infile, img, 0);
+			copy_to_img(infile, img, 0);
 			break;
 
 		case COPYRESIZED:
-			copy_to_gif(infile, img, 1);
+			copy_to_img(infile, img, 1);
 			break;
 
 		case TRANSPARENT:
@@ -759,6 +767,28 @@ START:	type = get_token(infile);
 			}
 			break;
 
+		case TYPE: 
+			imgtype = get_token(infile);
+			if(imgtype != JPEG && imgtype != PNG && imgtype != GIF){
+				fprintf(verbose_out,"Error: line %d: Image type must be jpeg, gif or png.\n",line_number);
+				exit(1);
+			}
+			break;
+
+		case QUALITY:
+			i=get_number(infile);
+			jp_quality = i;
+			if(imgtype != JPEG){
+				fprintf(verbose_out,"Error: line %d: Cannot use quality directive unless type is set to jpeg first.\n",line_number);
+				exit(1);
+			}
+			if(i < 0 || i > 95){
+				fprintf(verbose_out,"Line %d skipped: Image quality must be within the range 0-95. %d is not an acceptable value.\n",line_number, i);
+				exit(1);
+			}
+			
+			break;
+
 		case COMMENT:
 			sync_input(infile);
 			break;
@@ -767,7 +797,7 @@ START:	type = get_token(infile);
 			break;
 
 		case END:
-			gdImageGif(img,outfile);
+			gdImagePng(img,outfile);
 			fclose(outfile);
 			gdImageDestroy(img);
 			written_out = TRUE;
@@ -786,9 +816,15 @@ START:	type = get_token(infile);
 		}
 	}	while( ! finished );
 
-	/*  Write the gd to the GIF output file and exit */
+	/*  Write the gd to the image output file and exit */
 FINISH:	if(!written_out && img) {
-		gdImageGif(img,outfile);
+		if(imgtype == JPEG) {
+			gdImageJpeg(img,outfile,jp_quality);
+		} else if(imgtype == GIF) {
+			gdImageGif(img,outfile);
+		} else {
+			gdImagePng(img,outfile);
+		}
 		fclose(outfile);
 		gdImageDestroy(img);
 	}
@@ -883,16 +919,21 @@ int get_token(FILE *infile){
 		*end="end",
 		*ellipse="ellipse",
 		*colourchange="colourchange",
-		*rotate="rotate";
+		*rotate="rotate",
+		*type="type",
+		*quality="quality",
+		*gif="gif",
+		*png="png",
+		*jpeg="jpeg";
 
-	while(((c=getc(infile))!=EOF)&&(c!=' ')&&(c!='\n')&&(c!=',')&&(c!='='))
+	while(((c=getc(infile))!=EOF)&&(c!=' ')&&(c!='\n')&&(c!=',')&&(c!='=')&&(c!='\r'))
 	{
 		temp[i++]=c;
 		if(temp[0] == '#') break;
 	}
 	if (c == EOF) finished = TRUE;
 
-	if (c == '\n') {
+	if (c == '\n' || c == '\r') {
 		line_number++;
 		if (i == 0) { return EMPTY; }
 	} else if (c == EOF) {
@@ -1084,6 +1125,26 @@ int get_token(FILE *infile){
 		free(input_type);
 		return ROTATE;
 	}
+	if( strcmp(input_type, type) == 0){
+		free(input_type);
+		return TYPE;
+	}
+	if( strcmp(input_type, quality) == 0){
+		free(input_type);
+		return QUALITY;
+	}
+	if( strcmp(input_type, gif) == 0){
+		free(input_type);
+		return GIF;
+	}
+	if( strcmp(input_type, png) == 0){
+		free(input_type);
+		return PNG;
+	}
+	if( strcmp(input_type, jpeg) == 0){
+		free(input_type);
+		return JPEG;
+	}
 	free(input_type);
 	ungetc(c,infile);
 	return (int)NULL;
@@ -1173,16 +1234,16 @@ int get_colour(FILE *infile, gdImagePtr img){
 
 /******************************************************************************
 **
-**  copy_to_gif
+**  copy_to_img
 **
-**  Copies a gif to the current image. Location of gif and coordinates are
+**  Copies a img to the current image. Location of img and coordinates are
 **  specified on the input line.
 **
 **  Used by:
 **  copy, copyresized.
 **
 ******************************************************************************/
-void copy_to_gif(FILE *infile, gdImagePtr img, int resize){
+void copy_to_img(FILE *infile, gdImagePtr img, int resize){
 	int     i=0,arg[8];
 	char    *filename;
 	FILE	*img_to_copy;
@@ -1198,11 +1259,17 @@ void copy_to_gif(FILE *infile, gdImagePtr img, int resize){
 		arg[i]=get_number(infile);
 	}
 	if((img_to_copy = get_file(&line_number, &filename)) == NULL) {
-		fprintf(stderr,"Error: Cannot read existing GIF file \"%s\"\n",
+		fprintf(stderr,"Error: Cannot read existing image file \"%s\"\n",
 			filename);
 		exit(1);
 	}
-	img_file = gdImageCreateFromGif(img_to_copy);
+	if(imgtype == JPEG) {
+		img_file = gdImageCreateFromJpeg(img_to_copy);
+	} else if(imgtype == GIF) {
+		img_file = gdImageCreateFromGif(img_to_copy);
+	} else {
+		img_file = gdImageCreateFromPng(img_to_copy);
+	}
 	fclose(img_to_copy);
 
 	if( resize == 1 )
@@ -1257,9 +1324,9 @@ sync_input(FILE *infile)
 	int c = 0;
 
 	if( c == '\n' ) return;
-	while( ( (c=getc(infile)) != EOF ) && (c != '\n') ) ;
+	while( ( (c=getc(infile)) != EOF ) && (c != '\n') && (c != '\r')) ;
 	if( c == EOF ) finished = TRUE;
-	if( c == '\n' ) line_number++;
+	if( c == '\n' || c == '\r') line_number++;
 	return;
 }
 
@@ -1348,16 +1415,16 @@ gdImagePtr get_image(int type, int argc, char *argv[]){
 	int newtype;
 
 	if( type == EXISTING ) {
-		/*  fprintf(stderr,"Creating GIF from existing file:"); */
+		/*  fprintf(stderr,"Creating image from existing file:"); */
 		if ((in = get_file(&line_number, &filename)) == NULL ) {
-			fprintf(stderr,"Error: Cannot read existing GIF file \"%s\"\n",
+			fprintf(stderr,"Error: Cannot read existing image file \"%s\"\n",
 				filename);
 			exit(1);
 		}
 		else {
-			if(!quiet) fprintf(verbose_out,"Creating image from existing gif <%s>\n",
+			if(!quiet) fprintf(verbose_out,"Creating image from existing image <%s>\n",
 			    filename);
-			image = gdImageCreateFromGif(in);
+			image = gdImageCreateFromPng(in);
 			fclose(in);
 		}
 	}
@@ -1385,12 +1452,12 @@ gdImagePtr get_image(int type, int argc, char *argv[]){
 		{
 			if( output_file )
 			{
-				fprintf(verbose_out,"Creating new %d by %d gif, <%s>\n",
+				fprintf(verbose_out,"Creating new %d by %d image, <%s>\n",
 					num[1],num[2],output_file);
 			}
 			else
 			{
-				fprintf(verbose_out,"Creating new %d by %d gif\n",
+				fprintf(verbose_out,"Creating new %d by %d image\n",
 					num[1],num[2]);
 			}
 		}
@@ -1439,7 +1506,7 @@ my_newmem(size_t size)
 **
 **  Used by:
 **  get_image
-**  copy_to_gif
+**  copy_to_img
 **
 ******************************************************************************/
 
@@ -1489,7 +1556,7 @@ FILE *get_file(int *line_number, char **filename)
 		close(p[1]);
 		in = fdopen(p[0], "rb");
 #else
-	fprintf(stderr,"Error: Reading GIF from external command not supported on this Operating System\n");
+	fprintf(stderr,"Error: Reading image from external command not supported on this Operating System\n");
 	exit(1);
 #endif
 	} else {
